import time
from abc import ABC, abstractmethod
from typing import List, TYPE_CHECKING

from django.db.models import Q

from . import utils
from .aggregations import Aggregation, Annotation
from .exceptions import InvalidCommandError
from .filter import Filter
from .joins import JoinQuery


if TYPE_CHECKING:
    from .dgeq import GenericQuery



class Command(ABC):
    """Interface for commands."""
    regex = None
    
    
    @abstractmethod
    def __call__(self, query: 'GenericQuery', field: str, values: List[str]):
        pass



class Annotate(Command):
    """Annotations are made up of key value pairs delimited by a pipe `|`:
    `key:value|key:value`. Valid keys are :
    
    * `field` (`field=population`) - Name of the field used for to compute the
        annotation.
    * `to` (`to=population_avg`) - Name of the field where the result of the
        annotation will be displayed.|
    * `func` (`func=avg`) - Function to use for the annotation. Value must be
        key of `DGEQ_AGGREGATION_FUNCTION` dictionary.
    * `filters` (`filters=mountains.height=]1500'mountains.name=*Mount`) - Allow
        to add an apostrophe `'` separated list of filters. These filters
        support search modifiers.
    * `delayed` (`delayed=1`) - Whether the annotation will be applied before
        (`0`) or after (`1`) the filtering of the main query. Default is `0`.|
    
    You can declare multiple annotation using a comma `,` or with multiple
    declaration of `c:annotate`. Each annotation's `to` must be unique.
    
    Filters must be given related to the main query model, and not the model
    used for the annotation. So if you have a query on `country/` and want to
    annotate on `rivers` count your query must be :
    
    * `country/?c:annotate=field=rivers|to=rivers_count|func=count|filters=rivers.length=>2000`
    
    and not:
    
    * `country/?c:annotate=field=rivers|to=rivers_count|func=count|filters=length=>2000`
    
    *note the field used in `filters`*.
    
    Field created by annotations on `to` can be used in other commands, such as
    `Sort`, `Show` and even `Aggregate`. They can also be used in filters,
    making it possible to filter on rivers average length for instance.
    
    Created annotation will be appended to `query.annotations`.
    
    Also append the field use by annotation to `query.arbitrary_fields`. See
    `Annotate` for more information."""  # noqa
    
    regex = "^c:annotate$"
    
    
    def __call__(self, query: 'GenericQuery', field: str, values: List[str]):
        annotations = utils.split_list_strings(values)
        
        query.annotations = list()
        
        for a in annotations:
            a = Annotation.from_query_value(
                a, query.model, query.case, query.censor, query.arbitrary_fields
            )
            query.arbitrary_fields.add(a.to)
            query.queryset = a.apply(query.queryset)



class Aggregate(Command):
    """Create aggregations by parsing query's `c:aggregate` value.
    
    Aggregations are made up of key value pairs delimited by a pipe `|`:
    `key:value|key:value`. Valid keys are :

    * `field` (`field=population`) - Name of the field used for to compute the
        aggregation.
    * `to` (`to=population_avg`) - Name of the field where the result of the
        aggregation will be displayed.|
    * `func` (`func=avg`) - Function to use for the aggregation. Value must be
        key of `DGEQ_AGGREGATION_FUNCTION` dictionary.
    
    You can declare multiple aggregation using a comma `,` or with multiple
    declaration of `c:aggregate`. Each aggregation's `to` must be unique."""
    
    regex = "^c:aggregate$"
    
    
    def __call__(self, query: 'GenericQuery', field: str, values: List[str]):
        aggregations = utils.split_list_strings(values)
        aggregations = [
            Aggregation.from_query_value(
                a, query.model, query.censor, query.arbitrary_fields
            ).get()
            for a in aggregations
        ]
        
        query.result = {
            **query.result,
            **query.queryset.aggregate(**dict(aggregations)),
        }



class Case(Command):
    """Modify filters' case-sensitiveness by looking at query's `c:case` value.
    
    A value of 1 mean the filters are case-sensitive, 0 mean case-insensitive,
    default to 1 if `c:case` is absent."""
    
    regex = "^c:case$"
    
    
    def __call__(self, query: 'GenericQuery', field: str, values: List[str]):
        if not values[-1].isdigit():
            raise InvalidCommandError(
                'c:case', f"value must be either 0 or 1 (received '{values[-1]}')"
            )
        query.case = bool(int(values[-1]))



class Count(Command):
    """Add the number of object in the database matching the query.
    
    Count will be added in the key `count` in the result if `c:count` value is
    1, default to 0 if `c:count` is absent."""
    
    regex = "^c:count$"
    
    
    def __call__(self, query: 'GenericQuery', field: str, values: List[str]):
        if not values[-1].isdigit():
            raise InvalidCommandError(
                "c:count", f"value must be either 0 or 1 (received '{values[-1]}')"
            )
        
        if int(values[-1]):
            query.result['count'] = query.queryset.count()



class Distinct(Command):
    """Eliminate duplicate from the resulting rows.
    
    By default, a query will not eliminate duplicate rows. In practice, this is
    rarely a problem, because simple queries don’t introduce the possibility of
    duplicate result rows. However, if your query spans multiple tables, it’s
    possible to get duplicate results when a query is evaluated."""
    
    regex = "^c:distinct$"
    
    
    def __call__(self, query: 'GenericQuery', field: str, values: List[str]):
        if query.sliced:
            raise InvalidCommandError(
                "c:distinct", "cannot be used after 'c:start' or 'c:limit'"
            )
        
        if not values[-1].isdigit():
            raise InvalidCommandError(
                f'c:distinct$', f"value must be either 0 or 1 (received '{values[-1]}')"
            )
        if int(values[-1]):
            query.queryset = query.queryset.distinct()



class Evaluate(Command):
    """Evaluate the query, putting the result in the `rows` key of the result.
    
    Rows will not be computed if `c:evaluated` value is 0, default to 1 if
    `c:evaluate` is absent."""
    
    regex = "^c:evaluate$"
    
    
    def __call__(self, query: 'GenericQuery', field: str, values: List[str]):
        if not values[-1].isdigit():
            raise InvalidCommandError(
                f'c:evaluate', f"value must be either 0 or 1 (received '{values[-1]}')"
            )
        query.evaluated = int(values[-1])



class Filtering(Command):
    """Apply filters on the query by looking at every query's field not starting
    with `c:`.
    
    Filters are created by using the standard `field=value` of query strings.
    The `field` portion must correspond to a particular field of the queried
    model.
    
    You can also query on related model using a dot `.` notation. For instance,
    in order to find all the disaster in Kenya, on would use the following pair
    string :`country.name=Kenya`. Related field can be nested up to
    `DGEQ_MAX_FOREIGN_FIELD_DEPTH`.
    
    If you query directly on a related model, and not on one of its field
    (E.G. `country` instead of `country.name`), `dgeq` will use it's primary key
    (most of the time `id`).
    
    The value portion can be of different types :

    * `int` (`?field=2`) - Plain integer.
    * `float` (`?field=3.14`) - Use dot `.` as decimal separator.
    * `string` (`?field=string`) - Plain string.
    * `boolean` (`?field=1`) - Use non-negative integers (`0` is `False`,
            anything else is `True`).
    * `date` (`?field=2004-12-02T22:00`) - An ISO 8601 [1] compliant string.
    * `null` (`?field=`) - Do not put any value.
    
    A modifier may be used in front of the value portion of the query string to
    better filter the rows. Only one modifier may be used, the second modifier
    character would be considered to be part of the value.
    
    * `<` (`country/?population=<500000`) - Less than.
    * `[` (`country/?population=]500000`) - Less than or equal.
    * `>` (`country/?population=>500000`) - Greater than.
    * `]` (`country/?population=[500000`) - Greater than or equal.
    * `!` (`country/?population=!500000`) - Different than.
    * `^` (`country/?name=^United`) - Starts with a string.
    * `$` (`country/?name=$Islands`) - Ends with a string.
    * `*` (`country/?name=*istan`) - Contains a string.
    * `~` (`country/?name=~z`) - Do not contain a string.
    
    To combine search modifier, either use the comma `,` :
    `country/?population=[500000,]500000`, or create another `field=value`
    with the other modifier : `country/?population=[500000&population=]500000`
    
    Modifiers are combined with a logical `AND`. For instance to get all the
    country with their name starting with `United`, but that does not contains
    `States` :
    
    * `country/?name=^United,~States` or `country/?name=^United&name=~States`

    [1] https://en.wikipedia.org/wiki/ISO_8601
    """
    
    regex = r"^(?!c:).*$"
    
    
    def __call__(self, query: 'GenericQuery', field: str, values: List[str]):
        if query.sliced:
            raise InvalidCommandError(
                field, "You cannot filter on fields after 'c:start' or 'c:limit'"
            )
        
        values = utils.split_list_strings(values)
        filters = Q()
        for v in values:
            utils.check_field(field, query.model, query.censor, query.arbitrary_fields)
            filters &= Filter(field, v, query.case).get()
        
        query.queryset = query.queryset.filter(filters)



class Join(Command):
    """Allow to retrieve data of related models instead of only their PK.
    
     Joins are declared with the `c:join` field and is made up of key value
     pairs delimited by a pipe `|` : `key:value|key:value`. Valid keys are :

    * `field` (`field=region`) - Name of the field containing the related model.
    * `show` (`show=name'id`) - Only include the provided fields (multiple field
            names separated by an apostrophe `'`).
    * `hide` (`hide=id'countries`) - Include all field except the provided
            fields (multiple field names separated by an apostrophe `'`).
            Will be ignored if `show` is present.
    
    The following keys only make sense when `field` is either a
    `ManyToManyField`, its related field, or the related field of a
    `ForeignKey` :
    
    * `start` (`start=10`) - Start with the `Nth` object within the join (first
            is `0`). Default to `0`.
    * `limit` (`limit=20`) - Limit the number of object in the join, set to `0`
            to get all the objects (default to `0`).
    * `sort` (`sort=-area'id`) Sort the joined models by the given field
            (apostrophe `'` separated list)
    * `filters` (`filters=rivers=[1000'mountains=<3000`) Use `filters` to add an
            apostrophe `'` separated list of filters. These filters supports
            search modifiers.
    """
    
    regex = "^c:join$"
    
    
    def __call__(self, query: 'GenericQuery', field: str, values: List[str]):
        joins = utils.split_list_strings(values)
        
        # Create joins in the query
        query_joins = dict()
        for j in joins:
            j = JoinQuery.from_query_value(
                j, query.model, query.case, query.censor, query.arbitrary_fields,
            )
            query_joins[j.field] = j
        
        # Sort join by their number of related model so that no useless
        # intermediary joins are created
        sorted_query_join = sorted(
            query_joins.items(), key=lambda i: i[0].count("__")
        )
        
        for field, join in sorted_query_join:
            query.add_join(field, join, query.model, query.censor)
        
        # Prefetch joins
        for j in query.joins.values():
            query.queryset = j.prefetch(query.queryset)



class Show(Command):
    """Allow to choose which field to include or remove.
    
    By default, every field of the queried model are included in the result. Use
    the `c:show` field to declare a comma separated
    """
    
    regex = "^c:(show)|(hide)$"
    
    
    def __call__(self, query: 'GenericQuery', field: str, values: List[str]):
        # Reset fields so that previous hide or show does not affect this one
        query.fields = {
            f.get_accessor_name() if utils.is_reverse(f) else f.name
            for f in query.model._meta.get_fields()
        }
        
        fields = utils.split_list_strings(values)
        for f in fields:
            utils.check_field(f, query.model, query.censor, query.arbitrary_fields)
        
        if field == "c:show":
            query.fields = set(fields)
        else:
            query.fields |= set(query.arbitrary_fields)
            query.fields -= set(fields)



class Sort(Command):
    """Sort the resulting rows by the provided fields in `c:sort` field.
    
    Value must be a comma separated list of field. Prepend an hyphen `-` to use
    descending order on a specific field (E.G.: "-name").
    
    Sorting can be done on related field by using dot `.` notation.
    """
    
    regex = "^c:sort$"
    
    
    def __call__(self, query: 'GenericQuery', field: str, values: List[str]):
        if query.sliced:
            raise InvalidCommandError(
                "c:sort", "cannot be used after 'c:start' or 'c:limit'"
            )
        
        fields = utils.split_list_strings(values)
        
        for f in fields:
            utils.check_field(
                f if not f.startswith("-") else f[1:], query.model, query.censor,
                query.arbitrary_fields
            )
        
        query.queryset = query.queryset.order_by(*[f.replace(".", "__") for f in fields])



class Subset(Command):
    """Allow to retrieve only a subset of the result with the `c:start` and
    `c:limit` commands.
    
    Use `c:start=N` to include the results from the Nth row (first row is 0),
    default to 0.
    Use `c:limit=N` to include up to N rows. To include every row, set N to 0.
    Default to 1.
    """
    
    regex = "^c:((start)|(limit))$"
    
    
    def __call__(self, query: 'GenericQuery', field: str, values: List[str]):
        if field == "c:start":
            if not values[-1].isdigit():
                raise InvalidCommandError(
                    f'c:start', f"value must be a non-negative integers (received '{values[-1]}')")
        else:
            if not values[-1].isdigit():
                raise InvalidCommandError(
                    f'c:limit', f"value must be a non-negative integers (received '{values[-1]}')"
                )
        
        if field == "c:start":
            start = int(values[-1])
            query.queryset = query.queryset[start:]
            query.sliced = True
        else:
            limit = int(values[-1])
            if limit:
                query.queryset = query.queryset[:limit]
            query.sliced = True



class Time(Command):
    """Add the time taken by the server to compute your query.
    
    Time will be added in the key `time` in the result if `c:time` value is
    1, default to 0 if `c:time` is absent."""
    
    regex = "^c:time$"
    
    
    def __call__(self, query: 'GenericQuery', field: str, values: List[str]):
        if not values[-1].isdigit():
            raise InvalidCommandError(
                f'c:time', f"value must be either 0 or 1 (received '{values[-1]}')"
            )
        if int(values[-1]):
            query.result["time"] = time.time() - query._time  # noqa



import logging
import re
import time
from typing import Any, Dict, List, Type, Union

from django.conf import settings
from django.contrib.auth.models import AnonymousUser, User
from django.db import models
from django.http import QueryDict

from . import commands, utils
from .censor import Censor
from .exceptions import DgeqError
from .joins import JoinMixin


logger = logging.getLogger(__file__)

# List of commands used to construct the queryset. They are all called in the
# given order when calling the method `prepare()` and `evaluate()` of
# `GenericQuery`. This list can be changed with the setting `DGEQ_COMMANDS`.
DEFAULT_PRE_COMMANDS: List[commands.Command] = [
    commands.Case(),
    commands.Annotate(),
    commands.Filtering(),
    commands.Distinct(),
    commands.Sort(),
    commands.Subset(),
    commands.Join(),
    commands.Show(),
    commands.Aggregate(),
    commands.Evaluate(),
]
DGEQ_PRE_COMMANDS: List[commands.Command] = [
    utils.import_callable(p) for p in getattr(settings, "DGEQ_PRE_COMMANDS", DEFAULT_PRE_COMMANDS)
]

DEFAULT_POST_COMMANDS: List[commands.Command] = [
    commands.Count(),
    commands.Time(),
]
DGEQ_POST_COMMANDS: List[commands.Command] = [
    utils.import_callable(p) for p in getattr(settings, "DGEQ_POST_COMMANDS", DEFAULT_POST_COMMANDS)
]

QueryDictType = Union[QueryDict, Type[QueryDict]]



class GenericQuery(JoinMixin):
    """Main class of the `dgeq` module.
    
    `public_fields` and `private_fields` allow the definition of *private* and
    *public* field specific model. For each model, only the field defined in
    public, or not defined in private, will be present in the resulting rows.
    If both private and public field are defined for a same model, only the
    public definition will be used.
    One can also decide to use django's permission implementation. Dgeq will
    check for the `view_[model]` permission for each related field.
    Note that permission can be used in conjunction with private and public
    field.
    If the user manually try to access a private field (or is missing a
    permission), an error will be raised as if the field didn't exists.
    
    To use it in a view, you need to create an instance with the corresponding
    `Model` and the requests' `QueryDict` and execute its `evaluate()` method :
    
    ```python
    q = dgeq.GenericQuery(request.user, models.Continent, request.GET)
    result = q.evaluate()
    ```
    
    You can then modify the result as needed or just return it as a
    `JsonResponse` :
    
    ```python
    q = dgeq.GenericQuery(request.user, models.Continent, request.GET)
    result = q.evaluate()
    return JsonResponse(result)
    ```
    
    If you want to hide the `password` and `email` of `User` :
    
    ```python
    private = {User : ['password', 'email']}
    q = dgeq.GenericQuery(
        request.user, models.Continent, request.GET, private_fields=private
    )
    ```
    
    If you want to hide everything but the `username` of `User` :
    
    ```python
    public = {User : ['username']}
    q = dgeq.GenericQuery(
        request.user, models.Continent, request.GET, public_fields=public
    )
    ```
    
    """
    
    result: Dict[str, Any]
    
    
    def __init__(self, user: Union[User, AnonymousUser], model: Type[models.Model],
                 query_dict: QueryDictType, public_fields: utils.FieldMapping = None,
                 private_fields: utils.FieldMapping = None, use_permissions: bool = False):
        super().__init__()
        
        self._query_dict_list = list(query_dict.lists())
        self._time = time.time()
        
        self.model = model
        self.censor = Censor(user, public_fields, private_fields, use_permissions)
        self.fields = {
            f.get_accessor_name() if utils.is_reverse(f) else f.name
            for f in model._meta.get_fields()
        }  # noqa
        self.arbitrary_fields = set()
        self.queryset = self.model.objects.all()
        self.result = {'status': True}
        self.case = True
        self.evaluated = True
        self.sliced = False
    
    
    def _evaluate(self) -> List[Dict[str, Any]]:
        fields = set(self.fields)
        fields |= set(self.arbitrary_fields)
        fields = self.censor.censor(self.model, fields)
        
        fields, one_fields, many_fields = utils.split_related_field(
            self.model, fields, self.arbitrary_fields
        )
        queryset = self.queryset.select_related(
            *[f for f in one_fields if f not in self.joins.keys()]
        )
        queryset = queryset.prefetch_related(
            *[f for f in many_fields if f not in self.joins.keys()]
        )
        
        rows = list()
        for item in queryset:
            rows.append(utils.serialize_row(item, fields, one_fields, many_fields, self.joins))
        
        return rows
    
    
    def evaluate(self):
        """Execute commands in `DGEQ_PRE_COMMANDS` on each field matching the
        command's regex, then evaluate the queryset (adding the rows to
        `self.result` before executing command in DGEQ_POST_COMMANDS.
        
        Return `self.result`.
        """
        try:
            for field, lst in self._query_dict_list:
                matching_commands = (c for c in DGEQ_PRE_COMMANDS if re.match(c.regex, field))
                for command in matching_commands:
                    command(self, field, lst)
            
            if self.evaluated:
                self.result['rows'] = self._evaluate()
            
            for field, lst in self._query_dict_list:
                matching_commands = (c for c in DGEQ_POST_COMMANDS if re.match(c.regex, field))
                for command in matching_commands:
                    command(self, field, lst)
            
            result = self.result
        
        except DgeqError as e:
            result = {
                "status":  False,
                "message": str(e),
                "code":    e.code,
                **{a: getattr(e, a) for a in e.details}
            }
        
        except Exception as e:  # pragma: no cover
            logger.warning("Unknown error in dgeq:", exc_info=True)
            result = {
                "status":  False,
                "message": str(e),
                "code":    "UNKNOWN"
            }
        
        return result



from typing import Any, TYPE_CHECKING, Type, Union

from django.conf import settings
from django.db import models


if TYPE_CHECKING:
    from .censor import Censor

# Default max depth before a field lookup fail. Can be overridden in settings.py
MAX_FOREIGN_FIELD_DEPTH = getattr(settings, "DGEQ_MAX_FOREIGN_FIELD_DEPTH", 10)



class DgeqError(Exception):
    """Base exception for `dgeq` module.
    
    Exception inheriting `DgeqError` must declare attributes `code` containing
    an string containing only the character `[A-Z_]`. This code must be
    unique. Code `UNKNOWN` is reserved and must not be used.
    
    Such exceptions can also declare an attribute `details` containing a list
    of attribute's name that be output into the error payload."""
    
    code = "BASE_ERROR"
    details = []



class SearchModifierError(DgeqError):
    """Raised when using a wrong combination of search modifier and value."""
    
    code = "INVALID_SEARCH_MODIFIER"
    details = ['modifier', 'value', 'type']
    
    
    def __init__(self, modifier: str, value: Any):
        self.modifier = modifier
        self.value = value
        self.type = type(value).__name__
    
    
    def __str__(self):
        return (
            f"Search modifier '{self.modifier}' cannot be used on type "
            f"'{type(self.value).__name__}' "
            f"(type was extrapolated from value '{self.value}')"
        )



class UnknownFieldError(DgeqError):
    """Raised when an unknown field is used inside a `Filter` or in any
    command."""
    
    code = "UNKNOWN_FIELD"
    details = ['valid_fields', 'unknown']
    
    
    def __init__(self, model: Type[models.Model], unknown: str, censor: 'Censor'):
        include_hidden = getattr(settings, "DGEQ_INCLUDE_HIDDEN", False)
        
        self.model = model
        self.unknown = unknown
        self.valid_fields = [
            f.get_accessor_name() if f.auto_created and f.is_relation else f.name
            for f in model._meta.get_fields(include_hidden=include_hidden)
        ]
        self.valid_fields = list(censor.censor(model, self.valid_fields))
    
    
    def __str__(self):
        return (
            f"Unknown field '{self.unknown}' in table '{self.model.__name__}', valid fields are "
            f"{self.valid_fields}"
        )



class NotARelatedFieldError(DgeqError):
    """Raised when using a field that is not a related field."""
    
    code = "NOT_A_RELATED_FIELD"
    details = ['model_name', 'field', 'foreign_fields']
    
    
    def __init__(self, model: Union[Type[models.Model], models.Model], field: str,
                 censor: 'Censor'):
        include_hidden = getattr(settings, "DGEQ_INCLUDE_HIDDEN", False)
        
        self.model = model
        self.model_name = model.__name__
        self.field = field
        self.foreign_fields = [
            f.get_accessor_name() if f.auto_created else f.name
            for f in model._meta.get_fields(include_hidden=include_hidden) if f.is_relation
        ]
        self.foreign_fields = list(censor.censor(model, self.foreign_fields))
    
    
    def __str__(self):
        return (
            f"Field '{self.field}' in table '{self.model.__name__}, is neither a foreign key nor "
            f"a list of foreign key. Valid fields are {self.foreign_fields}"
        )



class FieldDepthError(DgeqError):
    """Raised if a field lookup exceed `MAX_FOREIGN_FIELD_DEPTH`."""
    
    code = "FIELD_DEPTH_ERROR"
    details = ['field', 'max_depth']
    
    
    def __init__(self, field: str):
        self.field = field
        self.max_depth = MAX_FOREIGN_FIELD_DEPTH
    
    
    def __str__(self):
        return (
            f"Field `{self.field}` exceed the allowed depth of related field of"
            f" {MAX_FOREIGN_FIELD_DEPTH}"
        )



class InvalidCommandError(DgeqError):
    """Raised when a commands is misused or its value is invalid."""
    
    code = "INVALID_COMMAND_ERROR"
    details = ['command']
    
    
    def __init__(self, command: str, message: str):
        self.command = command
        self.message = message
    
    
    def __str__(self):
        return f"Invalid command `{self.command}`: {self.message}"



from typing import Dict, Iterable, Type, Union

from django.conf import settings
from django.contrib.auth.models import AnonymousUser, User
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import FieldDoesNotExist
from django.db import models

from dgeq.exceptions import UnknownFieldError


# Fields always private
DEFAULT_PRIVATE_FIELDS = {
    User: ["password"]
}
DGEQ_PRIVATE_FIELDS = getattr(settings, "DGEQ_PRIVATE_FIELDS", DEFAULT_PRIVATE_FIELDS)



class Censor:
    """Allow to completely remove specific fields of specific models in the
    resulting rows.
    
    Censor instance take two dictionary mapping `Model` to a list of
    public / private fields.
    
    A field will be removed if:
    
    * `model` has defined public fields and `field` is not in the list.
    * `model` has defined private fields and `field` is in the list.
    * `field` if a related field, `self.use_permissions` is `True`, and the
       current `user` has not the *view* permission on the related model.
       
    If `model` has both explicitly defined public and private fields, private
    fields are ignored.
    """
    
    
    def __init__(self, user: Union[User, AnonymousUser],
                 public: Dict[Type[models.Model], Iterable[str]] = None,
                 private: Dict[Type[models.Model], Iterable[str]] = None,
                 use_permissions=False):
        self.user = user
        self.public = public or dict()
        self.private = private or dict()
        self.use_permissions = use_permissions
    
    
    def is_private(self, model: Type[models.Model], field: str) -> bool:
        """Return `True` if this `Model`'s `field` should be removed for the
        current user."""
        if self.use_permissions:
            try:
                field_instance = model._meta.get_field(
                    field if not field.endswith("_set") else field[:-4]
                )
            except FieldDoesNotExist:
                raise UnknownFieldError(model, field, self)
            if field_instance.is_relation:
                content_type = ContentType.objects.get_for_model(field_instance.remote_field.model)
                if not self.user.has_perm(f"{content_type.app_label}.view_{content_type.model}"):
                    return True
        
        if model in DGEQ_PRIVATE_FIELDS and field in DGEQ_PRIVATE_FIELDS[model]:
            return True
        
        if model in self.public:
            return field not in self.public[model]
        
        if model in self.private:
            return field in self.private[model]
        
        return False
    
    
    def is_public(self, model: Type[models.Model], field: str) -> bool:
        """Return `True` if this `Model`'s `field` should not be removed for the
        current user."""
        return not self.is_private(model, field)
    
    
    def censor(self, model: Type[models.Model], fields: Iterable[str]):
        """Return a censored set corresponding to the public field of
        `fields`."""
        return {f for f in fields if self.is_public(model, f)}



from datetime import datetime

from django.conf import settings
from django.db.models import Q, QuerySet

from . import types
from .exceptions import SearchModifierError
from .utils import import_callable


# Parsers for query string values. The whole list can be redefined with the
# `DGEQ_TYPE_PARSERS` setting. Parsers must ba a callable or a dotted path
# (E.G. `dgeq.parsers.none_parser`) to a callable which return either the parsed
# type of Ellipsis if no type were matched. If every parser return Ellipsis, the
# value will be interpreted as a string.
DEFAULT_TYPE_PARSERS = [
    types.none_parser,
    types.int_parser,
    types.float_parser,
    types.datetime_parser,
]
TYPE_PARSERS = [
    import_callable(p) for p in getattr(settings, "DGEQ_TYPE_PARSERS", DEFAULT_TYPE_PARSERS)
]

# Table giving which django's lookup function to use according to the search
# modifier and the type. New entry can be added with the `DGEQ_FILTERS_TABLE`
# setting. Default entry can be disable by setting their value to `None`.
#
# This effectively allow to add search modifier or to modify on by declaring a
# entry and disabling the old one corresponding to the same function.
#
# For strings, the value is a tuple with the first element being the lookup
# to use if the search is NOT case-sensitive.
DEFAULT_FILTERS_TABLE = {
    ('', int):         'exact',
    ('', str):         ('iexact', 'exact'),
    ('', datetime):    'exact',
    ('', type(None)):  'exact',
    
    ('!', int):        'exact',
    ('!', str):        ('iexact', 'exact'),
    ('!', datetime):   'exact',
    ('!', type(None)): 'exact',
    
    ('>', int):        'gt',
    ('>', float):      'gt',
    ('>', str):        ('gt', 'gt'),
    ('>', datetime):   'gt',
    
    ('[', int):        'gte',
    ('[', float):      'gte',
    ('[', str):        ('gte', 'gte'),
    ('[', datetime):   'gte',
    
    ('<', int):        'lt',
    ('<', float):      'lt',
    ('<', str):        ('lt', 'lt'),
    ('<', datetime):   'lt',
    
    (']', int):        'lte',
    (']', float):      'lte',
    (']', str):        ('lte', 'lte'),
    (']', datetime):   'lte',
    
    ('^', str):        ('istartswith', 'startswith'),
    ('$', str):        ('iendswith', 'endswith'),
    ('*', str):        ('icontains', 'contains'),
    ('~', str):        ('icontains', 'contains'),
}
FILTERS_TABLE = {
    **DEFAULT_FILTERS_TABLE,
    **getattr(settings, "DGEQ_FILTERS_TABLE", dict()),
}

# Compute search modifiers from FILTERS_TABLE
SEARCH_MODIFIERS = [i[0] for i in FILTERS_TABLE.keys()]

# Default behaviour is to use `queryset.filter()` when filtering. Search
# modifier in this list will use `queryset.exclude()` instead. This list can
# be changed with the setting `DGEQ_EXCLUDE_SEARCH_MODIFIER`
EXCLUDE_SEARCH_MODIFIERS = getattr(settings, 'DGEQ_EXCLUDE_SEARCH_MODIFIER', ['!', '~'])



class Filter:
    """Represent a search filter in a `GenericQuery`"""
    
    
    def __init__(self, field: str, value: str, case: bool):
        """Create a `Filter` from field-value pair of a query string.
        
        `case` indicate whether the filter is case-sensitive (`True`) or not.
        
        `value` must include the search modifier, if any.
        """
        # Extract optional modifier from value, if any
        if value and value[0] in SEARCH_MODIFIERS:
            modifier, value = value[0], value[1:]
        else:
            modifier = ""
        
        self.field = field.replace(".", "__")
        self.modifier = modifier
        for parser in TYPE_PARSERS:
            if (v := parser(value)) is not ...:
                self.value = v
                break
        else:
            self.value = value
        self.case = case
    
    
    def get(self) -> Q:
        """Return a `Q` object corresponding to this `Filter`."""
        try:
            lookup = FILTERS_TABLE[(self.modifier, type(self.value))]
            if isinstance(self.value, str):
                lookup = lookup[self.case]
        except KeyError:
            raise SearchModifierError(self.modifier, self.value)
        
        if self.modifier not in EXCLUDE_SEARCH_MODIFIERS:
            q = Q(**{self.field + "__" + lookup: self.value})
        else:
            q = ~Q(**{self.field + "__" + lookup: self.value})
        
        return q
    
    
    def __eq__(self, other: 'Filter'):
        return (
            self.value == self.value
            and self.field == self.field
            and self.modifier == self.modifier
        )
    
    
    def apply(self, queryset: QuerySet) -> QuerySet:
        """Create a new `QuerySet` by applying this filter to the given
        one."""
        return queryset.filter(self.get())



"""Setuptools entry point."""
import codecs
import os

from setuptools import setup


CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Internet',
    'Environment :: Web Environment',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.8',
    'Framework :: Django',
    'Framework :: Django :: 1.8',
    'Framework :: Django :: 1.10',
    'Framework :: Django :: 1.11',
    'Framework :: Django :: 2.0',
    'Framework :: Django :: 2.1',
    'Framework :: Django :: 2.2',
    'Framework :: Django :: 3.0',
    'Framework :: Django :: 3.1',
]

dirname = os.path.dirname(__file__)

short_description = (
    "Provide a query string API for construction of complex django QuerySet"
    " - https://dgeq.readthedocs.io/"
)
long_description = (codecs.open(os.path.join(dirname, 'README.md'), encoding='utf-8').read() + '\n'
                    + codecs.open(os.path.join(dirname, 'documentation/CHANGELOG.md'), encoding='utf-8').read())

setup(
    name='dgeq',
    version="0.3.5",
    description=short_description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Coumes Quentin',
    author_email='coumes.quentin@gmail.com',
    url='https://github.com/qcoumes/dgeq',
    packages=['dgeq'],
    install_requires=['django', 'python-dateutil'],
    classifiers=CLASSIFIERS,
)



from functools import reduce
from typing import Iterable, Tuple, Type

from django.conf import settings
from django.db import models
from django.db.models import QuerySet

from . import utils
from .censor import Censor
from .exceptions import InvalidCommandError, UnknownFieldError
from .filter import Filter


# Function used in aggregation and annotation. New entry can be added by with
# the setting `DGEQ_AGGREGATION_FUNCTION`. Existing one can be disabled by
# setting them to `None`.
DEFAULT_AGGREGATION_FUNCTION = {
    "max":    models.Max,
    "min":    models.Min,
    "avg":    models.Avg,
    "sum":    models.Sum,
    "stddev": models.StdDev,
    "var":    models.Variance,
    "count":  models.Count,
    "dcount": utils.DistinctCount,
}
AGGREGATION_FUNCTION = {
    **DEFAULT_AGGREGATION_FUNCTION,
    **{
        k: utils.import_callable(v)
        for k, v in getattr(settings, "DGEQ_AGGREGATION_FUNCTION", dict()).items()
    },
}



class Aggregation:
    """Represents an aggregation.
    
    Fields:
        * `field` (`str`) - Field used to compute the aggregation.
        * `to` (`str`) - Name of the field where the result of the aggregation
                will be stored.
        * `func` (`Type[models.Aggregate]`) - Function used for the aggregation.
    """
    
    
    def __init__(self, field: str, to: str, func: Type[models.Aggregate]):
        self.field = field
        self.to = to
        self.func = func
    
    
    @classmethod
    def from_query_value(cls, value: str, model: Type[models.Model], censor: Censor,
                         arbitrary_fields: Iterable[str] = ()) -> 'Aggregation':
        """Create an `Aggregation` from a 'c:aggregate` query string."""
        try:
            query_dict = utils.subquery_to_querydict(value)
        except ValueError as e:
            raise InvalidCommandError("c:aggregate", str(e))
        
        # Retrieve the field used to compute the aggregation
        if "field" not in query_dict:
            raise InvalidCommandError("c:aggregate", "'field' argument is missing")
        utils.check_field(query_dict["field"], model, censor, arbitrary_fields)
        field = query_dict["field"].replace(".", "__")
        
        # Retrieve the function used to compute the aggregation
        if "func" not in query_dict:
            raise InvalidCommandError("c:aggregate", "'func' argument is missing")
        if query_dict["func"] not in AGGREGATION_FUNCTION:
            raise InvalidCommandError(
                "c:aggregate",
                f"Unknown function '{query_dict['func']}', valid functions are : "
                f"{list(AGGREGATION_FUNCTION.keys())}"
            )
        func = AGGREGATION_FUNCTION[query_dict["func"]]
        
        # Retrieve the field where to put the computed aggregation
        if "to" not in query_dict:
            raise InvalidCommandError("c:aggregate", "'to' argument is missing")
        if not query_dict["to"].isidentifier():
            raise InvalidCommandError(
                "c:aggregate",
                f"'to' value isn't a valid identifier ('{query_dict['to']}'). Valid identifiers "
                f"can use uppercase and lowercase letters 'A' through 'Z', the underscore "
                f"'_' and (except for the first character) the digits '0' through '9'."
            )
        try:
            # Use check_fields to check that a field DOES NOT exists by
            # expecting the exception it raises.
            utils.check_field(query_dict["to"], model, censor, arbitrary_fields)
            raise InvalidCommandError(
                "c:aggregate", f"'to' value ('{query_dict['to']}') is a already used by a field"
            )
        except UnknownFieldError:
            pass
        to = query_dict["to"]
        
        return cls(field, to, func)
    
    
    def get(self) -> Tuple[str, models.Aggregate]:
        """Return a tuple (keyword, aggregation).
        
        Can be used to create kwargs for a call to `queryset.aggregate()` :
        ```python
        kwargs = dict([a.get() for a in aggregations])
        queryset.aggregate(**kwargs)
        ```
        """
        return self.to, self.func(self.field)



class Annotation:
    """Represents an annotation.
    
    Fields:
        * `field` (`str`) - Field used to compute the annotation.
        * `to` (`str`) - Name of the field where the result of the annotation
                will be stored.
        * `func` (`Type[models.Aggregate]`) - Function used for the annotation.
        * `filters` (``) - Filters use for the annotation.
        * `early` (`bool`) - Whether the annotation will be applied before
            (`True`) or after (`False`) the filtering of the main query. See [1]
            for more information. Default is False.
            
    [1] https://docs.djangoproject.com/en/3.1/topics/db/aggregation/#order-of-annotate-and-filter-clauses
    """  # noqa
    
    
    def __init__(self, field: str, to: str, func: Type[models.Aggregate],
                 filters: Iterable[Filter] = (), early: bool = False):
        self.field = field
        self.to = to
        self.func = func
        self.filters = filters
        self.early = early
    
    
    @classmethod
    def from_query_value(cls, value: str, model: Type[models.Model], case: bool, censor: Censor,
                         arbitrary_fields: Iterable[str] = ()) -> 'Annotation':
        """Create an `Annotation` from a 'c:annotate` query string."""
        try:
            query_dict = utils.subquery_to_querydict(value)
        except ValueError as e:
            raise InvalidCommandError("c:annotate", str(e))
        
        # Retrieve the field used to compute the annotation
        if "field" not in query_dict:
            raise InvalidCommandError("c:annotate", "'field' argument is missing")
        utils.check_field(query_dict["field"], model, censor, arbitrary_fields)
        field = query_dict["field"].replace(".", "__")
        
        # Retrieve the function used to compute the annotation
        if "func" not in query_dict:
            raise InvalidCommandError("c:annotate", "'func' argument is missing")
        if query_dict["func"] not in AGGREGATION_FUNCTION:
            raise InvalidCommandError(
                "c:annotate",
                f"Unknown function '{query_dict['func']}', valid functions are : "
                f"{list(AGGREGATION_FUNCTION.keys())}"
            )
        func = AGGREGATION_FUNCTION[query_dict["func"]]
        
        # Retrieve the field where to put the computed annotation
        if "to" not in query_dict:
            raise InvalidCommandError("c:annotate", "'to' argument is missing")
        if not query_dict["to"].isidentifier():
            raise InvalidCommandError(
                "c:annotate",
                f"'to' value isn't a valid identifier ('{query_dict['to']}'). Valid identifiers "
                f"can use uppercase and lowercase letters 'A' through 'Z', the underscore "
                f"'_' and (except for the first character) the digits '0' through '9'."
            )
        try:
            # Use check_fields to check that a field DOES NOT exists by
            # expecting the exception it raises.
            utils.check_field(query_dict["to"], model, censor, arbitrary_fields)
            raise InvalidCommandError(
                "c:annotate", f"'to' value ('{query_dict['to']}') is a already used by a field"
            )
        except UnknownFieldError:
            pass
        to = query_dict["to"]
        
        # Retrieve optional filters used for the annotation
        filters = list()
        for f in utils.split_list_strings(query_dict.getlist("filters"), "'"):
            kwarg = f.split('=', 1)
            if len(kwarg) < 2:
                raise InvalidCommandError(
                    "c:annotate", f"Filters must contains an equal '=', received '{kwarg[0]}'"
                )
            k, v = kwarg
            filters.append(Filter(k, v, case))
        
        # Check if this annotation must be delayed
        if not (early := query_dict.get("early", "0")) in ["0", "1"]:
            raise InvalidCommandError("c:annotate", "'early' argument must be '0' or '1'")
        early = bool(int(early))
        
        return cls(field, to, func, filters, early)
    
    
    def apply(self, queryset: QuerySet) -> QuerySet:
        """Return a new `QuerySet` with this `Annotation` applied."""
        if self.filters:
            kwargs = {
                self.to: self.func(
                    self.field, filter=reduce(lambda i, j: i & j, (f.get() for f in self.filters))
                )
            }
        else:
            kwargs = {self.to: self.func(self.field)}
        
        return queryset.annotate(**kwargs)



import operator
from collections import abc
from functools import reduce
from typing import Any, Callable, Dict, Iterable, List, Set, Tuple, Type, Union

from django.conf import settings
from django.contrib.auth.models import AnonymousUser, User
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models import Field
from django.http import QueryDict
from django.utils.module_loading import import_string

from .censor import Censor
from .exceptions import (FieldDepthError, MAX_FOREIGN_FIELD_DEPTH, NotARelatedFieldError,
                         UnknownFieldError)


# Separators use in subquery for some commands (like `c:annotate` or `c:join`)
# `SUBQUERY_SEP_FIELDS` is used to separate field/value(s) pairs
# `SUBQUERY_SEP_VALUES` is used to separate values in a same field
SUBQUERY_SEP_FIELDS = getattr(settings, 'DGEQ_SUBQUERY_SEP_FIELDS', '|')
SUBQUERY_SEP_VALUES = getattr(settings, 'DGEQ_SUBQUERY_SEP_VALUES', "'")

# Type corresponding to the Union of each field used for relations
ForeignField = Union[
    models.OneToOneField, models.OneToOneRel,
    models.ForeignKey, models.ManyToOneRel,
    models.ManyToManyField, models.ManyToManyRel
]

# Type mapping a Model to a subset of its field
FieldMapping = Dict[Type[models.Model], Iterable[str]]

# Type corresponding to a Callable or a dotted path to a Callable with its
# optional arguments
ImportableCallable = Union[
    Callable,  # Function or instance of a class declaring __call__
    str,  # Dotted path to a function, or class with no argument
    Tuple[str],  # Dotted path to a class with no argument in a tuple
    Tuple[str, Iterable[Any]],  # Dotted path to a class with args
    Tuple[str, Dict[str, Any]],  # Dotted path to a class with kwargs
    Tuple[str, Iterable[Any], Dict[str, Any]],  # Dotted path to a class with args and kwargs
]
# Fields containing a list of Foreign Keys.
MANY_FOREIGN_FIELD = (models.ManyToOneRel, models.ManyToManyField, models.ManyToManyRel)
# Fields containing only one Foreign Key
UNIQUE_FOREIGN_FIELD = (models.OneToOneRel, models.OneToOneField, models.ForeignKey)



def get_field(field: str, model: Type[models.Model]) -> models.Field:
    """Return the field named `field` of the given `Model`.
    
    Also work on reverse fields. The default name used by reverse fields
    (`[model]_set`) does not correspond to the actual name of the field when
    using `[model]._meta.get_field()`."""
    return model._meta.get_field(field if not field.endswith("_set") else field[:-4])



def _check_field(fields: List[str], current_model: Type[models.Model], censor: Censor,
                 arbitrary_fields: Iterable[str] = ()) -> Tuple[Type[models.Model], str]:
    """Recursively check fields.
    
    Wrapped by `check_field()`, see `check_field()` docstring for more
    information."""
    token, subfields = fields[0], fields[1:]
    
    if censor.is_private(current_model, token):
        raise UnknownFieldError(current_model, token, censor)
    
    try:
        # Will raise FieldDoesNotExists if `token` does not correspond to a
        # field in `current_model` or an arbitrary field
        for a in arbitrary_fields:
            if token == a:
                # Set relation for the subfield check below
                field = Field()
                field.is_relation = False
                break
        else:
            field = get_field(token, current_model)
        
        # If there's at least one subfield, change `current_model`
        # to the next one, checking if the field correspond to a relation
        if subfields:
            if not field.is_relation:
                raise NotARelatedFieldError(current_model, token, censor)
            current_model = field.remote_field.model
    except FieldDoesNotExist:
        raise UnknownFieldError(current_model, token, censor)
    
    # Recursively check subfields, if any
    if subfields:
        return _check_field(subfields, current_model, censor)
    
    return current_model, token



def check_field(field: str, model: Type[models.Model], censor: Censor,
                arbitrary_fields: Iterable[str] = (), sep=".") -> Tuple[Type[models.Model], str]:
    """Recursively check that a field exists foreign field.
    
    `arbitrary_fields` can be a list of string indicating arbitrary field added
    by some `QuerySet` method (like `annotate()` or `prefetch_related()`.
    
    Foreign field are separated by `sep`. For instance, if a model `Book`
    has a foreign key 'library' to a model `Library` which in turn has a foreign
    key `owner` to a model `Person`, one could do `library.owner.name` using
    dot `.` as `sep`.
    
    Additionally return a tuple (model : `models.Model`, field : `str`)
    corresponding to the last model and last field. For instance, using the
    same models as above, calling `check_field(Book, "library.owner.name")`
    would return `(Person, "name")`.
    
    Raise
        * `UnknownFieldError` if any of the field or foreign fields does not
           exists.
        * `FieldDepthError` if the depth of foreign field exceed
           `MAX_FOREIGN_FIELD_DEPTH`.
        * `NotAForeignFieldError` if a field used as a relation isn't a foreign
           field.
    """
    field_list = field.split(sep)
    max_depth = getattr(
        settings, "MAX_FOREIGN_FIELD_DEPTH", MAX_FOREIGN_FIELD_DEPTH
    )
    if len(field_list) >= max_depth:
        raise FieldDepthError(field)
    
    return _check_field(field_list, model, censor, arbitrary_fields)



def get_field_recursive(field: str, model: Type[models.Model], censor: Censor,
                        sep=".") -> Union[models.Field, ForeignField]:
    """Return the instance of `models.Field` corresponding to `field` inside
    `model`.
    
    Field can traverse relation using `sep`. For instance, if a model `Book`
    has a foreign key 'library' to a model `Library` which in turn has a foreign
    key `owner` to a model `Person`, one could do `library.owner.name` using
    dot `.` as `sep`.
    
    May raise the same exceptions as `check_field()`.
    """
    second_last_model, last_field_name = check_field(field, model, censor, sep=sep)
    return get_field(last_field_name, second_last_model)



def import_callable(o: ImportableCallable) -> Callable:
    """Take a dotted path to a `Callable`, and return the corresponding
    object.
    
    If `o` is already a function or an instance of a class declaring
    `__class__`, it is return as is.
    
    If `o` is a `str`, it must be a dotted path to a funtion or a class
    declaring `__call__()` and a constructor with no argument. The function or
    an instance of the class will be returned.
    
    If `o` is an `Iterable`, its first element must be a dotted path to a
    class declaring `__call__()`:
    
    * If its length is 2, its second element must be either an `Iterable` or a
    `Mapping` containing respectively the arguments or the keyworded arguments
    that will be given to the class's constructors.
    
    * If its length is greater than 2, the second element must be an `Iterable`
    and the third element a `Mapping` containing respectively the arguments and
    the keyworded arguments that will be given to the class's constructors.
    
    >>> import_callable(len)
    <built-in function len>
    >>> import_callable("functools.reduce")
    <built-in function reduce>
    >>> class Dummy:
    ...     def __init__(self, *args, **kwargs):  # noqa
    ...             self.args, self.kwargs = args, kwargs
    ...     def __call__(self):
    ...             print(self.args, self.kwargs)
    >>> import_callable("__main__.Dummy")()
    () {}
    >>> import_callable(("__main__.Dummy",))()
    () {}
    >>> import_callable(("__main__.Dummy", ("foo", "bar")))()
    ('foo', 'bar') {}
    >>> import_callable(("__main__.Dummy", {"foo": "bar"}))()
    () {'foo': 'bar'}
    >>> import_callable(("__main__.Dummy", ("foo", "bar"), {"foo": "bar"}))()
    ('foo', 'bar') {'foo': 'bar'}
    """
    if callable(o):
        if isinstance(o, type):
            raise ValueError("Given callable cannot be a class")
        return o
    
    if isinstance(o, str):
        imported = import_string(o)
        if not callable(imported):
            raise ValueError("Given path does not point to a callable")
        
        if not isinstance(imported, type):  # Is not a class
            return imported
        
        args = tuple()
        kwargs = dict()
    
    elif isinstance(o, abc.Iterable):
        o = list(o)
        args = tuple()
        kwargs = dict()
        imported = import_string(o[0])
        if not isinstance(imported, type):
            raise ValueError("First element of given iterable does not point to a class")
        
        if len(o) == 2:
            if isinstance(o[1], abc.Mapping):
                args = tuple()
                kwargs = o[1]
            elif isinstance(o[1], abc.Iterable):
                args = o[1]
                kwargs = dict()
            else:
                raise ValueError(
                    "Second element of given iterable must be either an iterable or a mapping if "
                    "its length is 2"
                )
        
        elif len(o) == 3:
            if not isinstance(o[1], abc.Iterable):
                raise ValueError(
                    "Second element of given iterable must be an iterable if its length is greater "
                    "than 2"
                )
            if not isinstance(o[2], abc.Mapping):
                raise ValueError(
                    "Second element of given iterable must be a mapping if its length is greater "
                    "than 2"
                )
            args = o[1]
            kwargs = o[2]
    
    else:
        raise ValueError("Given element is neither a Callable, a str nor an iterable")
    
    imported = imported(*args, **kwargs)
    if not callable(imported):
        raise ValueError("Given class does not declare __call__")
    
    return imported



def subquery_to_querydict(qs: str, fields_sep: str = SUBQUERY_SEP_FIELDS,
                          values_sep: str = SUBQUERY_SEP_VALUES) -> QueryDict:
    """Create a `QueryDict` out of a subquery string.
    
    Subquery strings are value of commands using different key/value pairs,
    such as `c:annotate` or `c:join`.
    
    `fields_sep` is used to separate field/value pairs.
    `values_sep` is used to separate values in a field.
    
    Raise `ValueError` if no `=` is found in a key/value pair:
    
    >>> subquery_to_querydict("field1=value^field_value_without_equal", "^")
    Traceback (most recent call last):
        ...
    ValueError: A key/value pair must contains an equal '=', received
    'field_value_without_equal'
    """
    query_dict = QueryDict(mutable=True)
    
    for kwarg in qs.split(fields_sep):
        kwarg = kwarg.split("=", 1)
        if len(kwarg) < 2:
            raise ValueError(f"A key/value pair must contains an equal '=', received '{kwarg[0]}'")
        
        k, values = kwarg
        for v in values.split(values_sep):
            query_dict.appendlist(k, v)
    
    return query_dict



def split_list_strings(lst: Iterable[str], sep: str = ',') -> List[str]:
    """Return a list of the words in each string of `lst`, using `sep` as
    delimiter.
    
    >>> split_list_strings(["one,two", "three", "four,five,six"], ",")
    ['one', 'two', 'three', 'four', 'five', 'six']"""
    return reduce(operator.add, (i.split(sep) for i in lst)) if lst else []



def is_reverse(field: models.Field) -> bool:
    """Return `True` if `field` is a reverse relation, `False` otherwise."""
    return field.auto_created and field.is_relation



def is_many(field: models.Field) -> bool:
    """Return `True` if `field` is a relation to multiple instance,
    `False` otherwise."""
    return type(field) in MANY_FOREIGN_FIELD



def is_one(field: models.Field) -> bool:
    """Return `True` if `field` is a relation to a unique instance,
    `False` otherwise."""
    return type(field) in UNIQUE_FOREIGN_FIELD



def split_related_field(model: Type[models.Model], fields: Iterable[str],
                        arbitrary_fields: Iterable[str] = ()
                        ) -> Tuple[Set[str], Set[str], Set[str]]:
    """Split the given `fields` of `model` into a tuple of iterables
    `(fields, one_fields, many_fields)`.
    
    `one_fields` contains fields that are either `ForeignKey`, `OneToOneField`
    or `OneToOneRel`.
    `many_fields` contains fields that are either `ManyToManyField`,
    `ManyToManyRel` or `ManyToOneRel`.
    `fields` contains all other fields.
    
    `arbitrary_fields` must be an iterable of the arbitrary fields added by
    `QuerySet`'s method like `annotation()`."""
    set_fields = set()
    one_fields = set()
    many_fields = set()
    
    for field_name in fields:
        if field_name in arbitrary_fields:
            set_fields.add(field_name)
            continue
        
        field = get_field(field_name, model)
        if is_one(field):
            one_fields.add(field_name)
        elif is_many(field):
            many_fields.add(field_name)
        else:
            set_fields.add(field_name)
    
    return set_fields, one_fields, many_fields



class DistinctCount(models.Count):
    """Wrap `models.Count(distinct=True)`."""
    
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, distinct=True, **kwargs)



def serialize_row(instance: models.Model, fields: Iterable[str] = (),
                  one_fields: Iterable[str] = (), many_fields: Iterable[str] = (),
                  joins: Dict = None) -> Dict[str, Any]:
    """Serialize an `instance` of `models.Model` according to the given fields
    and joins.
    
    See `split_related_field()` for the different field definition."""
    if joins is None:
        joins = dict()
    
    row = {f: getattr(instance, f) for f in fields}
    for f in one_fields:
        if f in joins.keys():
            row[f] = joins[f].fetch(instance)
        else:
            row[f] = getattr(instance, f)
            row[f] = None if row[f] is None else row[f].pk
    
    for f in many_fields:
        if f in joins.keys():
            row[f] = joins[f].fetch(instance)
        else:
            row[f] = list(map(lambda o: o.pk, getattr(instance, f).all()))
    return row



def serialize(user: [User, AnonymousUser], instance: models.Model,
              public_fields: FieldMapping = None, private_fields: FieldMapping = None,
              use_permissions: bool = False) -> Dict[str, Any]:
    """Serialize an `instance` the same way `dgeq.GenericQuery` would serialize
    a row."""
    model = instance.__class__
    include_hidden = getattr(settings, "DGEQ_INCLUDE_HIDDEN", False)
    fields = {
        f.get_accessor_name() if is_reverse(f) else f.name
        for f in model._meta.get_fields(include_hidden=include_hidden)
    }
    fields = Censor(user, public_fields, private_fields, use_permissions).censor(model, fields)
    fields, one_fields, many_fields = split_related_field(model, fields)
    
    row = {f: getattr(instance, f) for f in fields}
    for f in one_fields:
        row[f] = getattr(instance, f)
        row[f] = None if row[f] is None else row[f].pk
    for f in many_fields:
        row[f] = list(map(lambda o: o.pk, getattr(instance, f).all()))
    
    return row
